//
// this is the structural version of the line-drawing algorithm as described
// in the proposal.  note that a fsm may be grafted on top of this so that
// "mode" may be replaced by a reset which initializes the fsm to its initial
// state.
//
// fsm would look as follows:
//
// m0 --> m1 --> m2 --> m3 --+
//                      ^    |
//                      +----+
//
// with reset asserted at any time causing a cycle back to state m0.
// note that this may be encapsulated on top of the package module to 
// save work and enhance modularity.
//
// m0: load x1, x2
// m1: load y1, y2
// m2: delay cycle
// m3: work cycle.  continue until eoln is high.
//
// 19/08/97ajb
//
module package(in1, in2, clk, eoln, mode, xout, yout);
input [7:0] in1;
input [7:0] in2;
input clk;
output eoln;
input [1:0] mode;
output [7:0] xout, yout;


wire [15:0] rc;
wire vectored [15:0] rcr7,rcr8;
wire vectored [1:0]  r5bus,r6bus,r5out,r6out;

wire vectored [7:0] x2x1,y2y1,x1out,x1r5,x2out,x1x2,y1y2;
wire vectored [7:0] r7out,r8out,y1out,y2out,y1r6;

assign xout=x1out;	// doing a simple xout=x1out will *only*
assign yout=y1out;	// assign the [0] elements if you're not
			// defined in a wire or port statement first!

//
// decode fsm position (00 01 10 11) to discrete m-control bits
//
and #1(m3, mode[0], mode[1]);	// m3 detect
not #1(mode0bar, mode[0]);	// m2 detect
and #1(m2, mode[1],mode0bar);
not #1(mode1bar, mode[1]);	// m1 detect
and #1(m1, mode1bar, mode[0]);
nor #1(m0, mode[0], mode[1]);	// m0 detect

//
// left half of diagram
//
f3 modx1(x1out, in1, x1r5, x1out, m3, m2, m1, m0, rc[15], clk);
f0 modx2(x2out, in2, m0, clk);

subtractor88 modx2x1(zerox2x1, x2x1, x2out, x1out,cox0),
		modx1x2(zerox1x2, x1x2, x1out, x2out,cox);
adder82 modx1r5(x1r5, x1out, r5bus);
killstep modr5(r5out, cox, m1, clk);
bus modr5bus(r5bus, r5out, eolnbar);
f7 modr7(r7out, x2x1, x1x2, m1, cox, clk);

//
// right half of diagram
//
f4 mody1(y1out, in1, y1r6, y1out, m3, m2, m1, rc[15], clk);
f0 mody2(y2out, in2, m1, clk);
subtractor88 mody2y1(zeroy2y1, y2y1, y2out, y1out,coy0),
		mody1y2(zeroy1y2, y1y2, y1out, y2out,coy);
adder82 mody1r6(y1r6, y1out, r6bus);
killstep modr6(r6out, coy, m2, clk);
bus modr6bus(r6bus, r6out, eolnbar);
f7 modr8(r8out, y2y1, y1y2, m2, coy, clk);

//
// bottom portion 
//
f5 modrc(rc, zerox1x2 , rcr7, rcr8, m3, rc[15], clk);
adder168 modrcr7(rcr7, rc, r7out);
subtractor168 modrcr8(zerorcr8, rcr8, rc, r8out);

//
// eoln part
//
not #1 (eolnbar, eoln);
and #1(eoln, zerox1x2, zeroy2y1);

endmodule


//
// adds an 8-bit number to a sign-extended two bit one.
//
module adder82(o, i1, i2);
output [7:0] o;
input  [7:0] i1;
input  [1:0] i2;

wire   [1:8] c;

badd b[0:7] (c,o[0:7],i1[0:7],{i2[0],{7{i2[1]}}},{1'b0,c[1:7]});

endmodule


//
// adds a sixteen-bit number to a positive eight bit one.
//
module adder168(o, i1, i2);
output [15:0] o;
input  [15:0] i1;
input  [7:0] i2;

wire   [1:16] c;

badd b[0:15](			// rearranged named port example..
	.ci({1'b0,c[1:15]}),
	.a(i1[0:15]),
	.b({i2[0:7],8'h00}),
	.co(c), 
	.sum(o[0:15])
	);

endmodule

//
// single cell full adder
//
module badd(co, sum, a,b,ci);
output co,sum;
input a,b,ci;

xor #1	(sum,a^b,ci);
or #2	(co,a&b,b&ci,a&ci);

endmodule


//
// subtract one eight bit number from another
// use not(co) to input of r7/r8 muxes to make unsigned comparison
//
module subtractor88(z, o, i1, i2, co);
output z,co;
output [7:0] o;
input  [7:0] i1;
input  [7:0] i2;

wire [1:8]zer,car;

assign z=zer[8], co=car[8];

bsub b[1:8](car[1:8],o[0:7],i1[0:7],i2[0:7],{1'b1,car[1:7]},zer[1:8],
{1'b1,zer[1:7]});

endmodule


//
// subtract an eight-bit positive number from a sixteen bit one
//
module subtractor168(z, o, i1, i2);
output z;
output [15:0] o;
input  [15:0] i1;
input  [7:0] i2;

wire [1:16]zer,c;

assign z=zer[16];

bsub b[0:15](c[1:16],o[0:15],i1[0:15],{i2[0:7],8'h00},{1'b1,c[1:15]},
zer[1:16],{1'b1,zer[1:15]});

endmodule


//
// single cell "full subtractor"
//
module bsub(co, sum, a,b,ci, zo,zi);
output co,sum,zo;
input a,b,ci,zi;

not #1	(tb,b);

xor #1	(t1,a,tb),
	(sum,t1,ci);

or #2	(co,a&tb,tb&ci,a&ci);

not #1	(zdet,sum);
and #1  (zo,zi,zdet);

endmodule


//
// muxed register that accepts new input only on an m1 cycle
//
module f0(o, i, m, ck);
input [7:0] i;
input m,ck;
output [7:0] o;

mux1 m[7:0](o,i,{8{m}},{8{ck}});
 
endmodule
 

module mux1(o,i,m,ck);	//
input i,m,ck;		// this is for f0 module
output o;		//
 
and#1(t1,m,i);
not#1(t4,m);
and#1(t2,t4,o);
or#1(t3,t1,t2);
dff ga(o,ck,t3);

endmodule
 

//
// muxed register for x1 that only accepts as follows:
// m0			in1
// m3 & rc:15==0	in2
// else			in3
//
module f3(o,i1,i2,i3,m3,m2,m1,m0,s,c);

input [7:0] i1,i2,i3;
input m3,m2,m1,m0,s,c;
output [7:0] o;

mx3 m[7:0](o,i1,i2,i3,{8{m3}},{8{m2}},{8{m1}},{8{m0}},{8{s}},{8{c}});

endmodule


module mx3(o,in1,in2,in3, m3, m2, m1, m0, rc15,ck);
output o;
input in1,in2,in3,m3,m2,m1,m0,rc15,ck;

not #1(t1,rc15);

and #1	(t2,m0,in1),
	(t3,m1,in3),
	(t4,m2,in3),
	(t5,m3,rc15,in3),
	(t6,m3,t1,in2);

or #1	(t7,t2,t3,t4,t5,t6);
dff 	g03(o,ck,t7);

endmodule


//
// muxed register for y1 that only accepts as follows:
// m1			in1
// m3 & rc:15 == 1	in2
// (m3 & rc:15 == 0)	in3
//   | (m2)
// else			old y1
//
module f4(o,i1,i2,i3,m3,m2,m1,s,c);

input [7:0] i1,i2,i3;
input m3,m2,m1,s,c;
output [7:0] o;

mx4 m[7:0](o,i1,i2,i3,{8{m3}},{8{m2}},{8{m1}},{8{s}},{8{c}});

endmodule


module mx4(o,in1,in2,in3, m3, m2, m1, rc15,ck);
output o;
input in1,in2,in3,m3,m2,m1,rc15,ck;

not #1(t1,rc15);

and #1	(t2,m1,in1),
	(t3,m2,in3),
	(t4,m3,rc15,in2),
	(t5,m3,t1,in3);

or #1	(t6,t2,t3,t4,t5);

dff 	g04(o,ck,t6);
endmodule


//
// muxed register for rc that only accepts as following:
// m3 & rc:15 == 1	in2
// m3 & rc:15 == 0	in3
// else			in1 ($0000 or $8000)
//
module f5(o,i1,i2,i3,m3,s,c);

input i1;
input [15:0] i2,i3;
input m3,s,c;
output [15:0] o;

mx5a (o[15],i1,i2[15],i3[15],m3,s,c);
mx5b m[14:0] (o[14:0],i2[14:0],i3[14:0],{15{m3}},{15{s}},{15{c}});

endmodule


module mx5a(o,in1,in2,in3, m3, rc15,ck);
output o;
input in1,in2,in3,m3,rc15,ck;

not #1(t1,rc15);
not #1(t2,m3);

and #1	(t3,m3,rc15,in2),
	(t4,m3,t1,in3),
	(t5,t2,in1);

or #1	(t6,t3,t4,t5);

dff 	g04(o,ck,t6);

endmodule


module mx5b(o,in2,in3, m3, rc15,ck);
output o;
input in2,in3,m3,rc15,ck;

not #1(t1,rc15);

and #1	(t2,m3,rc15,in2),
	(t3,m3,t1,in3);

or #1	(t4,t2,t3);

dff 	g04(o,ck,t4);

endmodule


//
// muxed register for r7 and r8 that only accepts as following:
// m & s==0		i1
// m & s==1		i2
// else			previous contents
//
module f7(o, i1, i2, m, s, c);
output [7:0]o;
input  m,s,c;
input  [7:0] i1;
input  [7:0] i2;

mx2 m[7:0](o,i1,i2,{8{m}},{8{s}},{8{c}});

endmodule


module mx2(o,in1,in2,m,s,ck);
output o;
input in1,in2,m,s,ck;

not #1(t1,s);
not #1(t2,m);

and #1(t3,m,s,in2);
and #1(t4,m,t1,in1);
and #1(t5,t2,o);
or #1 (t6,t3,t4,t5);

dff g02(o,ck,t6);

endmodule


//
// selects whether r5 and r6 are zeroed out or not on eoln condition.
//
module bus(o, i, s);
output [1:0] o;
input [1:0] i;
input s;

assign #1 o=i&{2{s}};

endmodule


//
// the everpresent positive edge triggered d-ff
//
module dff(q,ck,d);
output q;
input ck,d;

nand#1  (q,x2,qbar),
	(qbar,q,x3),
	(x1,x4,x2),
	(x2,x1,ck),
	(x3,x2,ck,x4),
	(x4,x3,d);
endmodule


//
// stores either a +1 or -1 depending on whether zero out (i) is true
// 
//
module killstep(o,i,m,ck);
input ck, m, i;
output [1:0] o;
wire [1:0] o;
wire w;

assign o[0]=1;
assign o[1]=w;

mux1 m1(w,i,m,ck);

endmodule
